﻿# REBOTNIX FILTER AND UPLOAD REMONDIS DATA
import requests
import os
import sys
import datetime
import json
from os import listdir
from os.path import isfile, join
import base64
from operator import itemgetter
import shutil
import math
import hashlib 
from time import sleep
import cv2
import time 
import codecs

if "ROOT_PATH" in os.environ:
  sys.path.append(os.environ['ROOT_PATH'])

# Application version
version = "1.1.2"
print("RB-Filter-Upload for GUSTAV, Version:",version)

upload_success = 0

# setup a logger 
import libs.rb_logger as log
day = datetime.date.today().strftime("%d_%m_%Y")
if "ROOT_PATH" in os.environ:
  logdir = os.path.join(os.environ["ROOT_PATH"],"logs")
else:
  logdir = "./logs"
os.makedirs(logdir, exist_ok=True)
logfile = os.path.join(logdir ,"upload_" + day + ".log")

# give it a logger name and logfile
l = log.rb_logger().createlogger("upload",logfile)

c = 0
uc = 0
if "ROOT_PATH" in os.environ:
  cfg = open(os.path.join(os.environ['ROOT_PATH'],"cfg","main.json"), "r")
else:
  cfg = open("config.json","r")

text_cfg = cfg.read()
decoded_cfg=codecs.decode(text_cfg.encode(), 'utf-8-sig')
cfg_data = json.loads(decoded_cfg)

# DB information
carid = cfg_data["carid"]
url = 'https://datafleet-datafunction.azurewebsites.net/api/v1/messages/'+carid
header = {'Content-Type': 'application/json',
            'x-functions-key': 'zdQ4nICe3V3anSrHhxXVNgZ28dvb9tadDKm4FuMQWg/j1f1n38YdPQ=='}

# vehrkehrszeichenz katalog
if "ROOT_PATH" in os.environ:
  with open(os.path.join(os.environ['ROOT_PATH'],'cfg', 'vzkatalog.json')) as json_file:
      vzk = json.load(json_file)
else:
  with open('vzkatalog.json') as json_file:
      vzk = json.load(json_file)
vzlen = len(vzk)

def getBlackness(frame, hh, ww):
    #get brightness
    #convert to grayscale
    pixel_total = hh*ww
    gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    count = cv2.countNonZero(gray)
    blackness = (pixel_total-count)/pixel_total 
    return blackness

def wakeUp(content):
    response = requests.post(url,data=json.dumps(content), headers=header, verify=False)
    print("DB is awake... now sleep for 20 seconds")
    sleep(20)

def getMetaAll(filelist, debug, faileddir):
    all_dic = []
    fix_model_label = {"274_52":"274_20"}
    for index, filepath in enumerate(filelist):
        splits = os.path.basename(filepath)[:-4].split("_")
        fsize = os.path.getsize(filepath)

        #if file empty move to failed
        if fsize == 0:
            print(("filesize is 0 bytes, we move this!",filepath))
            if debug == False:
                shutil.move(filepath, os.path.join(faileddir, os.path.basename(filepath)))
            continue
        
        if len(splits) == 5:
            category = splits[0]
            ts = float(splits[1])
            lat = float(splits[2])
            lon = float(splits[3])
            acc = round(float(splits[4]),2)

        if len(splits) == 6:
            category = splits[0]+"_"+splits[1]
            ts = float(splits[2])
            lat = float(splits[3])
            lon = float(splits[4])
            acc = round(float(splits[5]),2)

        if category is "nosign":
            if debug == False:
                shutil.move(filepath, os.path.join(faileddir, os.path.basename(filepath)))
            continue

        if math.isnan(lat) or math.isnan(lon):
            if debug == False:
                shutil.move(filepath, os.path.join(faileddir, os.path.basename(filepath)))
            continue

        img = cv2.imread(filepath)
        height, width, channel = img.shape

        blackness = getBlackness(img, height, width)

        if blackness > 0.25:
            #print("Blackness: "+str(blackness))
            if debug == False:
                shutil.move(filepath, os.path.join(faileddir, os.path.basename(filepath)))
            continue

        
        with open(filepath, "rb") as image_file:
            _base64 = base64.b64encode(image_file.read())

        #fix label
        if category in fix_model_label:
            category = fix_model_label[category]

        rdc = {}
        rdc["width"]               = width
        rdc["height"]               = height
        rdc["containerversion"]    = '1.0.3'
        rdc["imageId"]             = os.path.basename(filepath)
        rdc["timestamp"]           = ts
        rdc["rawimage"]            = _base64.decode('utf-8')
        rdc["imagemd5"]            = hashlib.md5(_base64).hexdigest()
        rdc["imageformat"]         = 'jpg'

        # breitengard vs. laengengrad
        rdc["gps"]                 = [lat,lon]
        rdc["gps_accuracy"]        = 0.79
        rdc["modelid"]             = "trafficsign"
        rdc["modelversion"]        = "1.0"

        if category not in vzk:
            rdc["category"] = "unbekannt: " + category
            continue     
        else:
            rdc["category"]        = vzk[category]

        rdc["category"]            = vzk[category]
        rdc["classid"]             = category

        rdc["accuracy"]            = acc
        rdc["carid"]               = carid
        rdc["carident"]            = "az0000000000"

        #all_dic.append({"filepath":filepath, "category":category, "ts":ts, "lat":lat, "lon":lon, "accuracy":acc,
        #    "fsize":fsize, "_base64":_base64 })

        all_dic.append(rdc)

    #sort by timestamp
    #newlist = sorted(all_dic, key=itemgetter('ts')) 
    newlist = sorted(all_dic, key=itemgetter('timestamp')) 
    for index,el in enumerate(newlist):
        el["index"] = index

    return newlist


def sendFile(content, debug, rootdir, completedir, faileddir):
    global uc
    global c
    if "index" in content:
        del content["index"]
        del content["width"]
        del content["height"]

    response = None

    try:
        #Performs a POST on the specified url to get the service ticket
        response= requests.post(url,data=json.dumps(content), headers=header, verify=False)
        #response = "<Response [200]>"
    except Exception as e:
        l.info("Error in upload POST Call: " + str(e))
        print("Error in upload POST Call: " + str(e))
        pass

    #if everything ok move file
    if str(response)=="<Response [200]>":
        #move file
        if debug == False:
            uc = uc + 1
            c = 0
            l.info("succefull upload of file: "+str(content["imageId"]))
            print("succefull upload of file: "+str(content["imageId"]))
            if os.path.exists(os.path.join(rootdir, content["imageId"])):
                shutil.move(os.path.join(rootdir, content["imageId"]), os.path.join(completedir, content["imageId"]))
    else:
        l.error("Try "+str(c)+": failed upload of file: "+str(content["imageId"]+". Now retry to upload."))
        l.error("HTML ERROR CODE: "+str(response))
        print("Try "+str(c)+": failed upload of file: "+str(content["imageId"]+". Now retry to upload."))
        print("HTML ERROR CODE: "+str(response))
        sleep(0.05)
        c = c + 1
        if c > 50:
            text = "Error while uploading the information of file  "+str(content["imageId"])+". Moved to failed directory."
            l.info(text)
            print(text)
            shutil.move(os.path.join(rootdir, content["imageId"]), os.path.join(faileddir, content["imageId"]))
            c = 0
            return None
            #sendMail(text)
        sendFile(content, debug, rootdir, completedir, faileddir)
    
    sleep(0.05)


def copyAllFiles(rootdir, destdir):
    onlyfiles = [f for f in os.listdir(rootdir) if os.path.isfile(os.path.join(rootdir, f))]
    for f in onlyfiles:
        shutil.copy(os.path.join(rootdir,f), os.path.join(destdir,f))

def createDirLocally(path):
    #create local dir if it doesnt exist yet
    if not os.path.exists(path):
        os.makedirs(path)

#Rebotnix Preprocessor Class
class filter_upload_rebotnix:

    def __init__(self, rootdir, debug=True):
        self.rootdir = rootdir
        self.debug = debug

        #get date of today as string for new dest_dir
        now = datetime.datetime.now()
        now = str(now).split(" ")[0]
        kw = datetime.datetime.now().strftime("%V")
        if "ROOT_PATH" in os.environ:
          dest_dir = os.path.join(os.environ['ROOT_PATH'],"./images_kw"+kw)
        else:
          dest_dir = "./images_kw"+kw
        createDirLocally(dest_dir)

        # local folders where we move the files when we processed them
        self.completedir = os.path.join(dest_dir, "./uploadcomplete_"+now)
        self.queuedir = os.path.join(dest_dir, "./uploadqueue")
        self.restdir = os.path.join(dest_dir, "./rest")
        self.faileddir = os.path.join(dest_dir, "./failed")

        #create dirs
        createDirLocally(self.completedir)
        createDirLocally(self.restdir)
        createDirLocally(self.faileddir)
        createDirLocally(self.queuedir)

    ############################################
    #    2. Filter images locally + upload     #
    ############################################
    def filter(self):

        rootdir = self.rootdir

        # test calls
        im = 0
        maxcalls = 150
        starttime = 0
        endtime = 0
        now = datetime.datetime.now()

        # total time 
        totaltime = datetime.datetime.now()
        date1 = datetime.datetime.now()
        print("REBOTNIX DATA FILTER: " + str(date1))

        # date now
        start = datetime.datetime.now()

        #wake up T-Systems
        #if self.debug == False:
        #    print("Waking up T-Systems...")
        #    wakeUp("wake up content")
        #    print("done.")

        #get all cropped files
        onlyfiles = [os.path.abspath(join(rootdir, f)) for f in listdir(rootdir) if isfile(join(rootdir, f))] #os.path.abspath(join(rootdir, f))

        if len(onlyfiles) == 0:
            l.info("No images left to filter.")
            print("No images left to filter.")
            return 0

        print("Number of Images: " +str(len(onlyfiles)))
        l.info(("Number of Images: " +str(len(onlyfiles))+", in images directory."))
        print("create meta for all files...")
        l.info("create meta for all files...")
        #get meta for all files
        metaAll = getMetaAll(onlyfiles, self.debug, self.faileddir)
        print("done.")
        l.info("done creating meta for all files.")

        #now filter all files for duplicates
        frame_cache = {}
        frame_nr = 0
        final_frames = []

        print("Filter for duplicates...")
        l.info("filter for duplicates... ")
        for i, meta_A in enumerate(metaAll[:-1]):
            frame_nr = frame_nr + 1
            meta_B = metaAll[i+1]

            if frame_nr == 0 and meta_A["category"] not in frame_cache:
                frame_cache[meta_A["category"]] = {"count": 1, "category": meta_A["category"],"ts":meta_A["timestamp"],"index":meta_A["index"],"imageId":meta_A["imageId"], "time_distance": 0, "similarity":1, "frame_nr": frame_nr, "width":meta_A["width"], "height":meta_A["height"]}
                continue
            if meta_B["category"] in frame_cache:
                if frame_nr - frame_cache[meta_B["category"]]["frame_nr"] < 15:
                    frame_cache[meta_B["category"]]["count"] = frame_cache[meta_B["category"]]["count"] + 1
                    frame_cache[meta_B["category"]]["frame_nr"] = frame_nr
                    #check image size and take bigger
                    if meta_B["width"]*meta_B["height"] > frame_cache[meta_B["category"]]["width"]*frame_cache[meta_B["category"]]["height"]:
                        #print("bigger found")
                        frame_cache[meta_B["category"]]["width"] = meta_B["width"]
                        frame_cache[meta_B["category"]]["height"] = meta_B["height"]
                        frame_cache[meta_B["category"]]["index"] = meta_B["index"]
                    
                else:
                    if frame_cache[meta_B["category"]]["count"] >= 2:
                        final_frames.append(frame_cache[meta_B["category"]])
                    frame_cache[meta_B["category"]] = {"count": 1, "category": meta_A["category"],"ts":meta_A["timestamp"],"index":meta_A["index"],"imageId":meta_A["imageId"], "time_distance": 0, "similarity":1,"frame_nr": frame_nr, "width":meta_A["width"], "height":meta_A["height"]}
                        
            else:
                frame_cache[meta_B["category"]] = {"count": 1, "category": meta_A["category"],"ts":meta_A["timestamp"],"index":meta_A["index"],"imageId":meta_A["imageId"], "time_distance": 0, "similarity":1,"frame_nr": frame_nr, "width":meta_A["width"], "height":meta_A["height"]}

        for el in frame_cache.items():
            if el[1]["count"] >= 2:
                final_frames.append(el[1])
            
        print("Filtering done.")
        l.info("done filtering for duplicates.")

        print(str(len(final_frames))+" valid signs were detected.")
        l.info(str(len(final_frames))+" valid signs were detected.")

        #move valid crops to upload_queue
        for el in final_frames:
            if self.debug == False:
                shutil.move(os.path.join(rootdir, el["imageId"]), os.path.join(self.queuedir, el["imageId"]))
        
        #remove rest files to rest dir
        onlyfiles = [os.path.abspath(join(rootdir, f)) for f in listdir(rootdir) if isfile(join(rootdir, f))]
        for f in onlyfiles:
            if self.debug == False:
                if os.path.exists(f):
                    shutil.move(f, os.path.join(self.restdir, os.path.basename(f)))

        return len(final_frames)

    def upload(self):
        print("Starting the Upload...")
        filelist = [os.path.abspath(join(self.queuedir, f)) for f in listdir(self.queuedir) if isfile(join(self.queuedir, f))]
        alreadyUploaded = len([os.path.abspath(join(self.completedir, f)) for f in listdir(self.completedir) if isfile(join(self.completedir, f))])

        metaAll = getMetaAll(filelist, 0, self.faileddir)
        
        l.info(str(len(metaAll))+" images of "+ str(alreadyUploaded + len(metaAll)) +" in the upload queue.")
        print(str(len(metaAll))+" images of "+ str(alreadyUploaded + len(metaAll)) +" in the upload queue.")
        
        #now send information to DB
        for el in metaAll:
            if self.debug == False:
                sendFile(el, self.debug, self.queuedir, self.completedir, self.faileddir)            
        print("... Upload done.")

        if self.debug:
            print(str(uc + alreadyUploaded) + " of "+ str(alreadyUploaded + len(metaAll)) +" files were not uploaded, because we are in debug mode.")
            l.info(str(uc + alreadyUploaded) + " of "+ str(alreadyUploaded + len(metaAll)) +" files were not uploaded, because we are in debug mode.")
        else:
            print(str(uc + alreadyUploaded) + " of "+ str(alreadyUploaded + len(metaAll)) +" files were uploaded.")
            l.info(str(uc + alreadyUploaded) + " of "+ str(alreadyUploaded + len(metaAll)) +" files were uploaded succesfully.")

        return uc
